// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <util.h>
#include <true_false.h>
#include "../../dump_ctrl.h"
#include "../../complex_solver.h"
#include "../../const.h"
#include "../../light.h"
#include "functions.h"

//static double min_light_error=1e-10;


void light_free(struct light *in)
{
light_free_memory(in);
}

double light_cal_photon_density(struct light *in)
{
int i;
int ii;
double tot=0.0;

double H_tot=0.0;
double photons_tot=0.0;

for (ii=0;ii<in->points;ii++)
{
	tot=0.0;
	H_tot=0.0;
	photons_tot=0.0;
	for (i=0;i<in->lpoints;i++)
	{
		in->E_tot_r[i][ii]=in->Ep[i][ii]+in->En[i][ii];
		in->E_tot_i[i][ii]=in->Enz[i][ii]+in->Epz[i][ii];

		in->pointing_vector[i][ii]=0.5*epsilon0*cl*in->n[i][ii]*(pow(in->E_tot_r[i][ii],2.0)+pow(in->E_tot_i[i][ii],2.0));
		in->photons[i][ii]=in->pointing_vector[i][ii]*(in->l[i]/(hp*cl));
		in->photons_asb[i][ii]=in->photons[i][ii]*in->alpha[i][ii];

		double E=((hp*cl)/in->l[i])/q-in->Eg;

		//getchar();
		if (E>0.0)
		{
		in->H[i][ii]=E*q*in->photons_asb[i][ii];
		}else
		{
		in->H[i][ii]=0.0;
		}

		//printf("%d %d %le %le %le\n",i,ii,E,in->H[i][ii],in->photons_asb[i][ii]);
		photons_tot+=in->photons[i][ii]*in->dl;
		tot+=in->photons_asb[i][ii]*in->dl;
		H_tot+=in->H[i][ii]*in->dl;
		//printf("%le %le\n",E,in->l[i]);
		//getchar();

	}

in->Gn[ii]=tot;
in->Gp[ii]=tot;

in->H1d[ii]=H_tot;
in->photons_tot[ii]=photons_tot;

}


/*FILE *out;
out=fopen("./light_test.dat","w");
for (i=0;i<in->lpoints;i++)
{
	tot=0.0;
	for (ii=0;ii<in->points;ii++)
	{
		fprintf(out,"%le %le %le\n",in->l[i],in->x[ii],in->alpha[i][ii]*(in->sun_photons[i])*(exp(-in->alpha[i][ii]*(in->x[ii]))+exp(-2*in->alpha[i][ii]*in->x[in->points-1])*exp(in->alpha[i][ii]*in->x[ii])));
	}

in->G[ii]=tot;
}
fclose(out);*/



return 0.0;
}


void light_solve_optical_problem(struct light *in)
{
int i;

if (get_dump_status(dump_iodump)==TRUE) printf("light_solve_optical_problem\n");

	double Psun=in->Psun*pow(10.0,-in->ND);

	light_set_sun_power(in,Psun,in->laser_eff);

	if ((in->laser_eff==0)&&(in->Psun==0))
	{
		if (get_dump_status(dump_optics)==TRUE) printf("It's dark I know what the answer is\n");
		for (i=0;i<in->lpoints;i++)
		{
			memset(in->En[i], 0.0, in->points*sizeof(double));
			memset(in->Ep[i], 0.0, in->points*sizeof(double));
			memset(in->Enz[i], 0.0, in->points*sizeof(double));
			memset(in->Epz[i], 0.0, in->points*sizeof(double));
		}
	}else
	{
		light_solve_all(in);
	}

	light_cal_photon_density(in);

	light_dump(in);

	for (i=0;i<in->lpoints;i++)
	{
		light_dump_1d(in, i,"");
	}

}

int light_solve_lam_slice(struct light *in,int lam)
{
if (get_dump_status(dump_optics)==TRUE)
{
	char one[100];
	sprintf(one,"Solve light optical slice at %lf nm\n",in->l[lam]*1e9);
	//printf("%s\n",one);
	waveprint(one,in->l[lam]*1e9);
}
int i;

double complex Epl=0.0+0.0*I;
double complex Epc=0.0+0.0*I;
double complex Epr=0.0+0.0*I;

//double complex Enl=0.0+0.0*I;
double complex Enc=0.0+0.0*I;
double complex Enr=0.0+0.0*I;

double complex r0=0.0+0.0*I;
double complex r1=0.0+0.0*I;
double complex t0=0.0+0.0*I;
double complex t1=0.0+0.0*I;

//double yl=0.0;
//double yc=0.0;
//double yr=0.0;
double dyl=0.0;
//double dyr=0.0;
//double dyc=0.0;
double xl=0.0;
double xc=0.0;
//double xr=0.0;

int ittr=0;
int pos=0;
//double error=1000;
double l=in->l[lam];
//double nc;
double alpha_c;
double alpha_r;
double tot_error=0.0;
int quit=FALSE;
double test=FALSE;
double n0c;
double n0r;

do
{
	pos=0;

	for (i=0;i<in->points;i++)
	{
		if (i==0)
		{
			Epl=in->sun_E[lam]+0.0*I;
//			Enl=in->sun_E[lam]+0.0*I;		//don't care
			xl=in->x[i]-(in->x[i+1]-in->x[i]);


		}else
		{
			Epl=in->Ep[lam][i-1]+in->Epz[lam][i-1]*I;
//			Enl=in->En[lam][i-1]+in->Enz[lam][i-1]*I;
			xl=in->x[i-1];

		}

		if (i==(in->points-1))
		{
			Epr=1.0+0.0*I;
			Enr=1.0+0.0*I;
//			xr=in->x[i]+(in->x[i]-in->x[i-1]);
			r1=in->r[lam][i];
			t1=in->t[lam][i];
			n0r=in->n[lam][i];
			alpha_r=in->alpha[lam][i];
		}else
		{
			Epr=in->Ep[lam][i+1]+in->Epz[lam][i+1]*I;
			Enr=in->En[lam][i+1]+in->Enz[lam][i+1]*I;
//			xr=in->x[i+1];
			r1=in->r[lam][i+1];
			t1=in->t[lam][i+1];
			n0r=in->n[lam][i+1];
			alpha_r=in->alpha[lam][i+1];
		}

		xc=in->x[i];
		dyl=xc-xl;
		n0c=in->n[lam][i];
		alpha_c=in->alpha[lam][i];
//		dyr=xr-xc;
//		dyc=(dyl+dyr)/2.0;


		Epc=in->Ep[lam][i]+in->Epz[lam][i]*I;
		Enc=in->En[lam][i]+in->Enz[lam][i]*I;
		r0=in->r[lam][i];
		t0=in->t[lam][i];//in->alpha[i][ii])
		double complex nbar_c=n0c+alpha_c*(l/(pi*4.0))*I;
		double complex xi_c=((2*pi)/l)*nbar_c;

		double complex nbar_r=n0r+alpha_r*(l/(pi*4.0))*I;
		double complex xi_r=((2*pi)/l)*nbar_r;

		//printf("here t0=(%le+i%le)\n",creal(t0),cimag(t0));
	//printf("t=(%le+i%le) n0=(%le+i%le) n1=(%le+i%le)\n",creal(in->r[i][ii]),cimag(in->r[i][ii]),creal(n0),cimag(n0),creal(n1),cimag(n1));
		//printf("r1=(%le+i%le)\n",creal(r1),cimag(r1));
		//printf("t0=(%le+i%le)\n",creal(t0),cimag(t0));
		//printf("t1=(%le+i%le)\n",creal(t1),cimag(t1));

		//getchar();


		double complex fp=0.0;
		double complex fn=0.0;

		double complex pa=t0;//Epl
		double complex pb=-cexp(-xi_c*dyl*I);//Epc*
//		double complex pc=0.0;
		double complex pbn=-r0*cexp(xi_c*dyl*I);

//		double complex na=0.0;
		double complex nb=t1;//Enc
		double complex nc=-cexp(xi_r*dyl*I);//Enr*
		double complex nbp=-r1*cexp(-xi_r*dyl*I);


		if (i==0)
		{
			fp=-Epl;
			pa=0.0;
		}else
		{
			fp=0.0+0.0*I;
		}

		if (i==(in->points-1))
		{
			fn=0.0;//Enr*cexp(xi*dyl*I);
			nb=1.0;
			nc=1.0;
		}else
		{
			fn=0.0;
		}

		//printf("dy=%le\n",dyl);
		//printf("a(%le+i%le) b(%le+i%le) c(%le+i%le) = %le\n",creal(pa),cimag(pa),creal(pb),cimag(pb),creal(pc),cimag(pc),creal(fp),cimag(fp));
		if (ittr!=0)
		{
			double complex fEp=Epl*t0-Epc*cexp(-xi_c*dyl*I)-r0*Enc*cexp(xi_c*dyl*I);
			double complex fEn=Enc*t1-Epr*r1*cexp(-xi_c*dyl*I)-Enr*cexp(xi_c*dyl*I);
			if (i==in->points-1) fEn=Epc+Enc;
			tot_error+=fabs(creal(fEp))+fabs(cimag(fEp))+fabs(creal(fEn))+fabs(cimag(fEn));
			printf("%le\n",tot_error);
			//printf("%d %le errors=fEp(%le+i%le) fEn(%le+i%le) \n",i,tot_error,creal(fEp),cimag(fEp),creal(fEn),cimag(fEn));
			//printf("errors fEp %le \n",);
			//getchar();
		}

		if (i!=0)///////////
		{
			in->Ti[pos]=i;
			in->Tj[pos]=i-1;
			in->Tx[pos]=creal(pa);
			in->Txz[pos]=cimag(pa);
			pos++;
		}

		in->Ti[pos]=i;/////////
		in->Tj[pos]=i;
		in->Tx[pos]=creal(pb);
		in->Txz[pos]=cimag(pb);
		pos++;


		in->Ti[pos]=in->points+i;///////////
		in->Tj[pos]=in->points+i;
		in->Tx[pos]=creal(nb);
		in->Txz[pos]=cimag(nb);
		pos++;

		in->Ti[pos]=i;/////////
		in->Tj[pos]=in->points+i;
		in->Tx[pos]=creal(pbn);
		in->Txz[pos]=cimag(pbn);
		pos++;

		if (i!=in->points-1)//////////
		{
			in->Ti[pos]=in->points+i;
			in->Tj[pos]=i+1;
			in->Tx[pos]=creal(nbp);
			in->Txz[pos]=cimag(nbp);
			pos++;
		}

		if (i!=(in->points-1))///////////////
		{
			in->Ti[pos]=in->points+i;
			in->Tj[pos]=in->points+i+1;
			in->Tx[pos]=creal(nc);
			in->Txz[pos]=cimag(nc);
			pos++;
		}else/////////////////
		{
			in->Ti[pos]=in->points+i;
			in->Tj[pos]=i;
			in->Tx[pos]=creal(nc);
			in->Txz[pos]=cimag(nc);
			pos++;
		}

		in->b[i]=creal(fp);
		in->bz[i]=cimag(fp);

		in->b[in->points+i]=creal(fn);
		in->bz[in->points+i]=cimag(fn);

		//printf("%d %le %le %d %d\n",i,in->b[i],in->b[i+in->points],pos,in->N);

	}

	//printf("%d %d %d\n",in->N,pos,in->M);
	//complex_solver_print(in->M,in->N,in->Ti,in->Tj, in->Tx, in->Txz,in->b,in->bz);
	//getchar();

	if (pos!=in->N)
	{
		printf("not solving because number does not match%d %d\n",pos,in->N);
		exit(0);
	}

	complex_solver(in->M,in->N,in->Ti,in->Tj, in->Tx, in->Txz,in->b,in->bz);

	for (i=0;i<in->points;i++)
	{
		double update;

		update=in->b[i];
		in->Ep[lam][i]=update;

		update=in->bz[i];
		in->Epz[lam][i]=update;

		//printf("b=%le+i%le\n",in->b[i],in->bz[i]);

		update=in->b[in->points+i];
		in->En[lam][i]=update;

		update=in->bz[in->points+i];
		in->Enz[lam][i]=update;

	}

	ittr++;

if (test==TRUE)
{
	if (ittr>1) quit=TRUE;
	getchar();
}else
{
	quit=TRUE;
}
}while(quit==FALSE);

if (test==TRUE)
{
	printf("Error in optical model tot error=%le\n",tot_error);
}

//getchar();
return 0;
}
